#include <type_traits>
#include <vector>
#include <iostream>

template<typename F, typename T, typename R>
__global__ void map_impl(F f, R* dst, T* src) 
{
	dst[threadIdx.x] = f(src[threadIdx.x]);
}

template<typename F, typename T, typename R = typename std::result_of<F(T)>::type>
std::vector<R> map(F f, std::vector<T> const& v)
{
	const auto n = v.size();
	std::vector<R> r( n );
	
	T* bv = nullptr;
	R* br = nullptr;

	cudaMalloc( (void**)&bv, sizeof(T)*n ); 
	cudaMalloc( (void**)&br, sizeof(R)*n ); 
	cudaMemcpy( bv, v.data(), sizeof(T)*n, cudaMemcpyHostToDevice ); 
	
	dim3 dimBlock( n, 1 );
	dim3 dimGrid( 1, 1 );
	map_impl<<<dimGrid, dimBlock>>>(f, br, bv);

	cudaMemcpy( r.data(), br, sizeof(R)*n, cudaMemcpyDeviceToHost );

	cudaFree( bv );
	cudaFree( br );
	return r;
}

int main()
{
	std::vector<int>    A{1, 2, 3, 4, 5, 6, 7, 8, 9};

	double q = 100.0;
	auto f = [=] __host__ __device__ (int x){ return (x*x)/q; };
	
	auto res = map(f, A);

	for( auto x : res )
	{
		std::cout << x << "\n";
	}
}
