#include <CL/sycl.hpp>

#include <iostream>

#include <array>

constexpr cl::sycl::access::mode sycl_read = cl::sycl::access::mode::read;
constexpr cl::sycl::access::mode sycl_write = cl::sycl::access::mode::write;

/* This is the class used to name the kernel for the runtime.
 * This must be done when the kernel is expressed as a lambda. */
template <typename T> class MapKernel;

template <typename F, typename T, size_t n>
void map_kernel(F f, std::array<T, n>& arr_dst, std::array<T, n> const& arr_src)
{
  cl::sycl::queue deviceQueue;
  cl::sycl::range<1> numOfItems{n};
  cl::sycl::buffer<T, 1> b_dst(arr_dst.data(), numOfItems);
  cl::sycl::buffer<T, 1> b_src(arr_src.data(), numOfItems);

  deviceQueue.submit([&](cl::sycl::handler &cgh)
  {
    auto src = b_src.template get_access<cl::sycl::access::mode::read>(cgh);
    auto dst = b_dst.template get_access<cl::sycl::access::mode::write>(cgh);
    cgh.parallel_for<class MapKernel<T>>(numOfItems, [=](cl::sycl::id<1> idx){ dst[idx] = f(src[idx]); });
  });
}

int main()
{
   std::array<float, 4> A = {{1, 2, 3, 4}};
   for(int i=0; i<4; ++i)
   {
     std::cout << A[i] << " ";
   } std::cout << "\n";
   
   std::array<float, 4> B;
   float c = 5.0f;
   map_kernel([=](float x){ return c*x; }, B, A);
   for(int i=0; i<4; ++i)
   {
     std::cout << B[i] << " ";
   } std::cout << "\n";
   return 0;
}
