#include <CL/sycl.hpp>

#include <iostream>

#include <vector>

constexpr cl::sycl::access::mode sycl_read = cl::sycl::access::mode::read;
constexpr cl::sycl::access::mode sycl_write = cl::sycl::access::mode::write;

/* This is the class used to name the kernel for the runtime.
 * This must be done when the kernel is expressed as a lambda. */
template <typename T> class ReduceKernel;

template <typename F, typename T>
T reduce_kernel(F f, std::vector<T> const& arr_src)
{
  auto n = arr_src.size();
  std::vector<T> arr_dst(n/8);
  {
	  cl::sycl::queue deviceQueue;
	  cl::sycl::range<1> nSrc{n};
	  cl::sycl::range<1> nDst{n/8};

	  cl::sycl::buffer<T, 1> b_src(arr_src.data(), nSrc);
	  cl::sycl::buffer<T, 1> b_dst(arr_dst.data(), nDst);

	  deviceQueue.submit([&](cl::sycl::handler &cgh)
	  {
	    auto src = b_src.template get_access<cl::sycl::access::mode::read>(cgh);
	    auto dst = b_dst.template get_access<cl::sycl::access::mode::discard_write>(cgh);
	    cgh.parallel_for<class ReduceKernel<T>>(nDst, [=](cl::sycl::id<1> idx)
	    {

		T acc = src[idx*8];
		for(int k=1; k<8; ++k)
		{
			acc = f(acc, src[idx*8+k]);
		}
		dst[idx] = acc;
	    });
	  });


  }

  T acc = arr_dst[0];
  for(size_t i=1; i<arr_dst.size(); ++i){ acc = f(acc, arr_dst[i]); }
  return acc;
}

int main()
{
   std::vector<double>    A(1024);
   std::fill(A.begin(), A.end(), 2.0);
   A[42] = 1.22;
   
   auto res = reduce_kernel([=](double x, double y){ return (x < y ? x : y); }, A);
   
   std::cout << "reduce result is: " << res << "\n";
   return 0;
}
