#include <type_traits>
#include <vector>
#include <iostream>

template<typename F, typename T>
__global__ void reduce_impl(F f, T* dst, T* src ) 
{
	auto i = threadIdx.x;

	T acc = src[i*8];
	for(int k=1; k<8; ++k)
	{
		acc = f(acc, src[i*8+k]);
	}
	dst[i] = acc;
}

template<typename F, typename T>
T reduce(F f, std::vector<T> const& v)
{
	const auto n = v.size();
	const auto nr = n / 8;
	std::vector<T> res( nr );
	
	T* bv = nullptr;
	T* br = nullptr;

	cudaMalloc( (void**)&bv, sizeof(T)*n ); 
	cudaMalloc( (void**)&br, sizeof(T)*nr ); 
	cudaMemcpy( bv, v.data(), sizeof(T)*n, cudaMemcpyHostToDevice ); 
	
	dim3 dimBlock( nr, 1 );
	dim3 dimGrid( 1, 1 );
	reduce_impl<<<dimGrid, dimBlock>>>(f, br, bv);

	cudaMemcpy( res.data(), br, sizeof(T)*nr, cudaMemcpyDeviceToHost );

	T acc = res[0];
	for(int i=1; i<res.size(); ++i){ acc = f(acc, res[i]); }

	cudaFree( bv );
	cudaFree( br );
	return acc;
}

int main()
{
	std::vector<double>    A(1024);
	std::fill(A.begin(), A.end(), 2.56);
	A[42] = 1.22;

	auto f = [] __host__ __device__ (double x, double y){ return (x < y ? x : y); };
	
	auto res = reduce(f, A);

	std::cout << "reduce result is: " << res << "\n";
}
